#include <iostream>

#include <math.h>
#ifdef __APPLE__
#include <OpenGL/OpenGL.h>
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cstdlib>

#define REPULSION_ACCEL 0.005
#define TARGET_ACCEL 0.0005
#define RANDOM_ACCEL 0.05

#include "flock.h"
#include "misc.h"
#ifndef TARGET_SHAPE
#define TARGET_SHAPE
#include "targetshape.h"
#endif

// initialize the flock with random locations, directions, sizes, and flaps
void flock::initialize() {
	flocksize = NUMFLOCK;
	for (int i=0; i<flocksize; i++) {
		birdflock[i].set_loc(randlim(-50,50),
							 randlim(-50, 50), 
							 randlim(-90, -110));

		birdflock[i].set_speed(randlim(-0.25,0.25), 
							   randlim(-0.25,0.25), 
							   randlim(-0.25,0.25));
		
		// sizes is good for a flock of 400
		birdflock[i].set_size(randlim(0.5, 1));

		// start each bird flapping at a random point in the flap
		birdflock[i].set_flaptimer(rand());
	}
	// assign a default target
	target[0] = 0;
	target[1] = 0;
	target[2] = -100.0;
}

// update each member of the flock, as the flock seeks a common target
void flock::update() {
	
	calcdistances();
	
	for (int i=0; i<flocksize; i++) {
		birdflock[i].updateloc();
		birdflock[i].flap();
		updatemember(i, target, 
					REPULSION_ACCEL, TARGET_ACCEL, RANDOM_ACCEL);
	}
}

// update a member of the flock, as the flock tries to make a shape
void flock::makeshape(targetshape myshape) {
	
	calcdistances();
	float mytarget[3];
	
	for(int i=0; i<flocksize; i++) {
		birdflock[i].updateloc();
		birdflock[i].flap();
		
		// figure out the target for this particular bird
		myshape.get_point_loc(i, mytarget);
		updatemember(i, mytarget, 
					REPULSION_ACCEL, TARGET_ACCEL*10, RANDOM_ACCEL);
	}
}				

/* update an individual flock member:
- Repel by the other flock members
- Accelerate towards the target
- Accelerate a bit randomly
*/
void flock::updatemember(int i, float target[], 
						float repulsion_accel,
						float target_accel,
						float random_accel) {
	// these floats store the accelerations due to other birds
	float accel_x = 0;
	float accel_y = 0;
	float accel_z = 0;
	
	// be repelled by the other birds
	for (int j=0; j<flocksize; j++) {
		if (i != j) {
			accel_x += distances[i][j][1] / distances[i][j][0];
			accel_y += distances[i][j][2]/ distances[i][j][0];
			accel_z += distances[i][j][3] / distances[i][j][0];
		}
	}
	birdflock[i].accelerate(accel_x, accel_y, accel_z, repulsion_accel);
	
	// accelerate towards the target
	birdflock[i].accelerate(target[0] - birdflock[i].get_x(),
							target[1] - birdflock[i].get_y(),
							target[2] - birdflock[i].get_z(),
							target_accel);
						
	// and a bit of random acceleration
	birdflock[i].accelerate(randlim(-1, 1), randlim(-1, 1), randlim(-1, 1), 
							random_accel);
	birdflock[i].apply_topspeed();
}

// reassigns the target of the entire flock
void flock::set_target(float tx, float ty, float tz) {
	target[0] = tx;
	target[1] = ty;
	target[2] = tz;
}

// draw each member of the flock
void flock::draw() {
	for (int i=0; i<flocksize; i++) {
		birdflock[i].draw();	
	}
}

// draw the bird in color
void flock::draw(float r, float g, float b) {
	for (int i=0; i<flocksize; i++) {
		birdflock[i].draw(r, g, b);	
	}
}

// draw the bird as a sphere
void flock::drawspheres() {
	for (int i=0; i<flocksize; i++) {
		birdflock[i].drawsphere();	
	}
}

// draw the target as a green ball
void flock::drawtarget() {
	glPushMatrix();
    glTranslatef(target[0], target[1], target[2]);
    glColor3f(0, 1, 0);
    glutSolidSphere(1, 16, 16);
    glPopMatrix();
}


// calculate the squared distance between each pair of birds,
// and saves normalized separation vectors between each pair of birds
void flock::calcdistances() {
	float temp;
	
	for (int i=0; i<flocksize; i++) {
		for (int j=0; j<i; j++) {
			distances[i][j][1] = birdflock[i].get_x() - birdflock[j].get_x();
			distances[i][j][2] = birdflock[i].get_y() - birdflock[j].get_y();
			distances[i][j][3] = birdflock[i].get_z() - birdflock[j].get_z();
			distances[i][j][0] =	pow(distances[i][j][1],2) +
									pow(distances[i][j][2],2) +
									pow(distances[i][j][3],2);
			
			distances[j][i][0] = distances[i][j][0];
		
			temp = sqrt(distances[i][j][0]);
		
			// normalize the separation vectors, and save it for the second
			// bird of the pair (inverted)
			for (int k=1; k<4; k++) {
				distances[i][j][k] = distances[i][j][k] / temp;
				distances[j][i][k] = -distances[i][j][k];
			}
		}
	}
}
